from .. import GMQLDataset, GDataframe
import os, shutil
from glob import glob
from ... import get_python_manager, get_remote_manager
from . import MetaLoaderFile, RegLoaderFile, MemoryLoader, Loader
from ...FileManagment.TempFileManager import get_unique_identifier, get_new_dataset_tmp_folder


def materialize(datasets):
    """ Multiple materializations. Enables the user to specify a set of GMQLDataset to be materialized.
    The engine will perform all the materializations at the same time, if an output path is provided,
    while will perform each operation separately if the output_path is not specified.

    :param datasets: it can be a list of GMQLDataset or a dictionary {'output_path' : GMQLDataset}
    :return: a list of GDataframe or a dictionary {'output_path' : GDataframe}
    """
    result = None
    if isinstance(datasets, dict):
        result = dict()
        for output_path in datasets.keys():
            dataset = datasets[output_path]
            if not isinstance(dataset, GMQLDataset.GMQLDataset):
                raise TypeError("The values of the dictionary must be GMQLDataset."
                                " {} was given".format(type(dataset)))
            gframe = dataset.materialize(output_path)
            result[output_path] = gframe
    elif isinstance(datasets, list):
        result = []
        for dataset in datasets:
            if not isinstance(dataset, GMQLDataset.GMQLDataset):
                raise TypeError("The values of the list must be GMQLDataset."
                                " {} was given".format(type(dataset)))
            gframe = dataset.materialize()
            result.append(gframe)
    else:
        raise TypeError("The input must be a dictionary of a list. "
                        "{} was given".format(type(datasets)))
    return result


def materialize_local(id, output_path=None):
    pmg = get_python_manager()

    if output_path is not None:
        # check that the folder does not exists
        if os.path.isdir(output_path):
            shutil.rmtree(output_path)

        pmg.materialize(id, output_path)
        pmg.execute()

        # taking in memory the data structure
        real_path = os.path.join(output_path, 'exp')

        remove_side_effects(real_path)
        # metadata
        meta = MetaLoaderFile.load_meta_from_path(real_path)
        # region data
        regs = RegLoaderFile.load_reg_from_path(real_path)
    else:
        # We load the structure directly from the memory
        collected = pmg.collect(id)
        regs = MemoryLoader.load_regions(collected)
        meta = MemoryLoader.load_metadata(collected)

    result = GDataframe.GDataframe(regs=regs, meta=meta)
    return result


def remove_side_effects(path):
    for f in glob(os.path.join(path, ".*")) + glob(os.path.join(path, "_*")):
        os.remove(f)


def materialize_remote(id, output_name=None, download_path=None, all_load=True):
    pmg = get_python_manager()
    if not isinstance(output_name, str):
        output_name = get_unique_identifier()
    pmg.materialize(id, output_name)
    remote_manager = get_remote_manager()
    if (download_path is None) and all_load:
        download_path = get_new_dataset_tmp_folder()
    result = remote_manager.execute_remote_all(output_path=download_path)
    # pmg.getServer().clearMaterializationList()
    if len(result) == 1:  # TODO: change this!!!
        path = result[0]
        return Loader.load_from_path(local_path=path, all_load=all_load)



import os

user_path = os.path.expanduser("~")
USER_FOLDER_NAME = ".pygmql"


def initialize_user_folder():
    if not os.path.isdir(get_user_dir()):
        os.makedirs(get_user_dir())


def get_user_dir():
    return os.path.join(user_path, USER_FOLDER_NAME)


import xml.etree.ElementTree as ET
import requests
from requests.exceptions import ConnectionError
from xml.dom import minidom
from pkg_resources import resource_filename
import os
from tqdm import tqdm
from glob import glob

CHUNK_SIZE = 5 * 1024 * 1024  # 5 MB


class DependencyManager:
    def __init__(self):
        self.resources_path = resource_filename("gmql", "resources")
        self.dependency_file_path = os.path.join(self.resources_path, "dependencies.xml")
        self.repo_name, self.repo_url, self.deps = self._parse_dependency_file(self.dependency_file_path)
        self.backend_info_file = os.path.join(self.resources_path, "dependencies_info.xml")
        backend_info = None
        if os.path.isfile(self.backend_info_file):
            backend_info = self._parse_dependency_info(self.backend_info_file)
        self.backend_info = backend_info
        backend_jar_path = glob(os.path.join(self.resources_path, "*.jar"))
        backend_jar_path = backend_jar_path[0] if len(backend_jar_path) == 1 else None
        self.backend_jar_path = backend_jar_path

    def is_backend_present(self):
        return self.backend_info is not None

    def is_connection_on(self):
        pass

    @staticmethod
    def _parse_dependency_file(path):
        tree = ET.parse(path)
        root = tree.getroot()

        repository = root.find("repository")
        repo_name = repository.find("name").text
        repo_url = repository.find("url").text

        if repo_url[-1] == '/':
            repo_url = repo_url[:-1]

        deps = []
        dependencies = root.find("dependencies")
        if dependencies is not None:
            for d in dependencies:
                groupId = d.find("groupId").text  # mandatory
                artifactId = d.find("artifactId").text  # mandatory
                version = d.find("version").text  # mandatory
                dd = {
                    'groupId': groupId,
                    "artifactId": artifactId,
                    "version": version
                }
                classifier = d.find("classifier")  # optional
                if classifier is not None:
                    classifier = classifier.text
                    dd['classifier'] = classifier
                deps.append(dd)
        return repo_name, repo_url, deps

    def _parse_dependency_info(self, path):
        tree = ET.parse(path)
        return self.__parse_dependency_info_from_tree(tree)

    def _parse_dependency_info_fromstring(self, s):
        tree = ET.ElementTree(ET.fromstring(s))
        return self.__parse_dependency_info_from_tree(tree)

    @staticmethod
    def __parse_dependency_info_from_tree(tree):
        root = tree.getroot()
        data = root.find("data")
        res = {}
        for d in data:
            tag = d.tag
            text = d.text
            try:
                n = int(text)
                text = n
            except Exception:
                pass
            res[tag] = text
        return res

    def resolve_dependencies(self):
        first_part_url = "/".join(self.repo_url.split("/")[:3]) + "/"
        query_url = first_part_url + "/service/local/artifact/maven/resolve?"
        for d in self.deps:
            query_url += "g={}&".format(d['groupId'])
            query_url += "a={}&".format(d['artifactId'])
            query_url += "v={}&".format(d['version'])
            query_url += "r={}".format(self.repo_name)
            if "classifier" in d.keys():
                query_url += "&c={}".format(d['classifier'])
            try:
                resp_text = requests.get(query_url).text
            except ConnectionError:
                if self.is_backend_present():
                    return self.backend_jar_path
                else:
                    raise ValueError("Unable to connect to repository to retrieve GMQL backend. Check your connection")
            resp = self._parse_dependency_info_fromstring(resp_text)
            location = self.repo_url + resp['repositoryPath']
            output_path = os.path.join(self.resources_path, resp['repositoryPath'].split("/")[-1])
            if not self.is_backend_present():
                # there is no backend (first start)
                self._download_backend(location, output_path)
                self._save_dependency(resp_text)
            elif self.repo_name == 'snapshots' and \
                    self.backend_info['snapshot'] == 'true':
                # we have a snapshot backend and we are pulling a snapshot
                current_timestamp = self.backend_info['snapshotTimeStamp']
                retrieved_timestamp = resp['snapshotTimeStamp']
                if current_timestamp < retrieved_timestamp:
                    # we are using an outdated backend
                    self._delete_current_backend()
                    self._download_backend(location, output_path)
                    self._save_dependency(resp_text)
            else:
                raise NotImplementedError("Need to implement the backend download in the case"
                                          " of releases!!!!")
            return output_path

    def _save_dependency(self, resp):
        resp_nice = minidom.parseString(resp)
        resp_nice = resp_nice.toprettyxml()
        with open(os.path.join(self.resources_path, "dependencies_info.xml"), "w") as f:
            f.write(resp_nice)

    @staticmethod
    def _download_backend(location, output_path):
        r = requests.get(location, stream=True)
        total_size = int(r.headers.get("content-length", 0))
        with open(output_path, "wb") as f:
            for data in tqdm(r.iter_content(chunk_size=CHUNK_SIZE), total=total_size/CHUNK_SIZE, unit="B",
                             unit_scale=True):
                f.write(data)

    def _delete_current_backend(self):
        # search for the only jar file in the resources path
        os.remove(self.backend_jar_path)
        os.remove(self.backend_info_file)



reg_fixed_fileds = ['chr', 'start', 'stop', 'strand']

chr_aliases = ['chr', 'chrom', 'chromosome']
start_aliases = ['start', 'left']
stop_aliases = ['stop', 'right', 'end']
strand_aliases = ['str', 'strand']
id_sample_aliases = ['id_sample', 'sample_id', 'sample']

operator_correspondences = {
    'LT': 'GT',
    'GT': 'LT',
    'LTE': 'GTE',
    'GTE': 'LTE',
    'EQ': 'EQ',
    'NOTEQ': 'NOTEQ'
}


def _get_opposite_operator(operator):
    return operator_correspondences[operator]



from setuptools import setup, find_packages
import os
import sys


with open(os.path.join(".", "gmql", "resources", "version"), "r") as f_ver:
    version = f_ver.read().strip()

with open(os.path.join(".", "gmql", "resources", "github_url"), "r") as f_ver:
    github_url = f_ver.read().strip()

# Checking the numpy installation
try:
    import numpy
except ImportError:
    sys.exit("install requires: 'numpy'.\n"
             "Use 'pip install numpy' first")


setup(name='gmql',
      version=version,
      description='Python library for GMQL',
      long_description="PyGMQL is a python module that enables the user "
                       "to perform operation on genomic data in a scalable way.",
      url=github_url,
      author='Luca Nanni',
      author_email='lucananni93dev@gmail.com',
      # maintainer="Mustafa Anil Tuncel",
      # maintainer_email="tuncel.manil@gmail.com",
      license='MIT',
      packages=find_packages(exclude=("gmql.ml",)),     # temporary
      scripts=['scripts/pygmql_login.py', 'scripts/pygmql_win.bat', 'scripts/pygmql.sh'],
      install_requires=['pandas', 'tqdm', 'numpy', 'py4j', 'requests', 'requests-toolbelt', 'strconv',
                        # 'sklearn', 'pyclustering', 'matplotlib', 'scipy', 'wordcloud', 'fancyimpute'   # temporary
                        ],
      classifiers=[
            'Development Status :: 3 - Alpha',
            'Intended Audience :: Science/Research',
            'Topic :: Scientific/Engineering :: Bio-Informatics',
             'Topic :: Scientific/Engineering :: Information Analysis',
            'Programming Language :: Python :: 3'
      ],
      keywords="genomics big data pandas python",
      python_requires='>=3',
      include_package_data=True,
      zip_safe=False)



from ..parsers import get_parsing_function
import pandas as pd
from tqdm import tqdm
import logging
# from concurrent.futures import ThreadPoolExecutor
import math
from ..DataStructures import reg_fixed_fileds
from multiprocessing import pool, cpu_count
from functools import partial
from ... import none, Some

# Loading strategy
loading_strategy = "single_core"


def load_regions(collected_result):
    # get the number of regions
    n_regions = collected_result.getNumberOfRegions()
    # get the string delimiters
    regions_delimiter = collected_result.REGIONS_DELIMITER()
    values_delimiter = collected_result.VALUES_DELIMITER()
    # get how the strings are structured
    names, types = get_schema(collected_result)
    result = []

    from ... import is_progress_enabled
    if loading_strategy == 'single_core':
        # get the full string
        regions_string = collected_result.getRegionsAsString(none())
        if regions_string:
            # convert to list of strings
            regions_string = regions_string.split(regions_delimiter)
            iterator = map(lambda x: string_to_dictionary(x, values_delimiter, names, types),
                           tqdm(regions_string, disable=not is_progress_enabled()))
            result.extend(iterator)
    elif loading_strategy == 'multi_core':
        # number of divisions
        divisions = 10
        chunk_size = math.ceil(n_regions / divisions)
        p = pool.Pool(min(4, cpu_count()))

        std_partial = partial(string_to_dictionary, values_delimiter=values_delimiter,
                              names=names, types=types)

        for _ in tqdm(range(divisions), disable=not is_progress_enabled()):
            # get the full string
            regions_string = collected_result.getRegionsAsString(Some(chunk_size))
            if regions_string:
                # convert to list of strings
                regions_string = regions_string.split(regions_delimiter)

                iterator = p.map(std_partial, regions_string)
                result.extend(iterator)

        p.close()
    else:
        raise ValueError("Unknown loading mode ({})".format(loading_strategy))

    columns = reg_fixed_fileds + names
    if len(result) > 0:
        df = pd.DataFrame.from_dict(result)
        df = df.set_index(keys="id_sample", drop=True)
        df = df[columns]  # for ordering the columns
    else:
        df = pd.DataFrame(columns=columns)
        df.index.name = "id_sample"
    return df


def string_to_dictionary(region, values_delimiter, names, types):
    elements = region.split(values_delimiter)
    # print(elements)
    d = dict()
    d['id_sample'] = int(elements[0])
    d['chr'] = elements[1]
    d['start'] = int(elements[2])
    d['stop'] = int(elements[3])
    d['strand'] = elements[4]

    for i in range(0, len(names)):
        name = names[i]
        fun = types[i]
        element = elements[5 + i]
        if element != 'null':
            d[name] = fun(element)
        else:
            d[name] = None
    return d


def to_dictionary(region, names, types):
    d = dict()
    d['id_sample'] = int(region[0])
    d['chr'] = region[1]
    d['start'] = int(region[2])
    d['stop'] = int(region[3])
    d['strand'] = region[4]

    for i in range(0, len(names)):
        name = names[i]
        fun = types[i]
        d[name] = fun(region[5 + i])

    return d


def load_metadata(collected_result):
    meta_Java = collected_result.getMetadata()
    meta_list = []
    for meta in meta_Java:
        id = int(meta[0])
        name = meta[1]
        value = meta[2]
        meta_list.append({'id_sample' : id, name: value})

    if len(meta_list) > 0:
        df = pd.DataFrame.from_dict(meta_list)
        columns = df.columns

        # grouping by 'id_sample'
        g = df.groupby('id_sample')

        result_df = pd.DataFrame()
        from ... import __disable_progress
        for col in tqdm(columns, total=len(columns), disable=__disable_progress):
            if col != 'id_sample':
                result_df[col] = g[col].apply(to_list)
    else:
        result_df = pd.DataFrame()
        result_df.index.name = "id_sample"
    return result_df


def to_list(x):
    l = list(x)
    l = [a for a in l if not pd.isnull(a)]
    return l


def get_schema(collected_result):
    schema_Java = collected_result.getSchema()
    # schema_Java = [[name, type], [name, type], ...]
    names = []
    types = []
    for field in schema_Java:
        name = field[0]
        type_string = field[1]
        fun = get_parsing_function(type_string)

        names.append(name)
        types.append(fun)
    return names, types



from .storers import FrameToGMQL
import pandas as pd
from .loaders import Loader
from .DataStructures import reg_fixed_fileds, \
    strand_aliases, stop_aliases, start_aliases, chr_aliases, id_sample_aliases
import numpy as np
from ..FileManagment import TempFileManager
# from ..ml.genometric_space import GenometricSpace

chr_types = [object, str]
start_types = [int, np.int, np.int8, np.int16, np.int32, np.int64]
stop_types = start_types
strand_types = [object, str]
id_sample_types = [object, str] + start_types

default_id_sample = "sample"


class GDataframe:
    """ Class holding the result of a materialization of a GMQLDataset.
    It is composed by two data structures:
    
        - A table with the *region* data
        - A table with the *metadata* corresponding to the regions
    """
    def __init__(self, regs=None, meta=None):
        if (regs is None) and (meta is None):
            raise ValueError("At least one of meta or regs must not be None")
        if (regs is not None) and (not isinstance(regs, pd.DataFrame)):
            raise TypeError("regs: expected pandas Dataframe, got {}".format(type(regs)))
        if (meta is not None) and (not isinstance(meta, pd.DataFrame)):
            raise TypeError("meta: expected pandas Dataframe, got {}".format(type(meta)))

        if regs is None:
            index = meta.index
            regs = pd.DataFrame(index=index)

        if meta is None:
            meta = empty_meta(regs)

        self.regs = regs
        self.meta = meta

    # def to_genomic_space(self):
    #     """ Translates the GDataframe to the Genomic Space data structure
    #
    #     :return: a GenometricSpace object
    #     """
    #     return GenometricSpace.from_memory(self.regs, self.meta)

    def to_dataset_files(self, local_path=None, remote_path=None):
        """ Save the GDataframe to a local or remote location

        :param local_path: a local path to the folder in which the data must be saved
        :param remote_path: a remote dataset name that wants to be used for these data
        :return: None
        """
        return FrameToGMQL.to_dataset_files(self, path_local=local_path, path_remote=remote_path)

    def to_GMQLDataset(self, local_path=None, remote_path=None):
        """ Converts the GDataframe in a GMQLDataset for later local or remote computation

        :return: a GMQLDataset
        """
        local = None
        remote = None
        if (local_path is None) and (remote_path is None):
            # get a temporary path
            local = TempFileManager.get_new_dataset_tmp_folder()
        if local_path is not None:
            local = local_path
        if remote_path is not None:
            remote = remote_path
        self.to_dataset_files(local, remote)

        if local is not None:
            return Loader.load_from_path(local_path=local)
        elif remote is not None:
            raise NotImplementedError("The remote loading is not implemented yet!")

    def _normalize_metadata(self):
        meta = self.meta
        meta = meta.apply(_normalize_column)
        return GDataframe(regs=self.regs, meta=meta)
    

def _normalize_column(column):
    lengths = column.map(len)
    if len(list(filter(lambda x: x > 1, lengths))) == 0:
        new_column = column.map(lambda x: x[0] if len(x) > 0 else np.nan)
    else:
        new_column = column
    return new_column


def from_pandas(regs, meta=None, chr_name=None, start_name=None, stop_name=None,
                strand_name=None, sample_name=None):
    """ Creates a GDataframe from a pandas dataframe of region and a pandas dataframe of metadata

    :param regs: a pandas Dataframe of regions that is coherent with the GMQL data model
    :param meta: (optional) a pandas Dataframe of metadata that is coherent with the regions
    :param chr_name: (optional) which column of :attr:`~.regs` is the chromosome
    :param start_name: (optional) which column of :attr:`~.regs` is the start
    :param stop_name: (optional) which column of :attr:`~.regs` is the stop
    :param strand_name: (optional) which column of :attr:`~.regs` is the strand
    :param sample_name: (optional) which column of :attr:`~.regs` represents the sample name
           of that region. If nothing is provided, all the region will be put in a single sample.
    :return: a GDataframe
    """
    regs = check_regs(regs, chr_name, start_name, stop_name, strand_name, sample_name)
    regs = to_gmql_regions(regs)
    if meta is not None:
        if not check_meta(meta, regs):
            raise ValueError("Error. Meta dataframe is not GMQL standard")
    else:
        meta = empty_meta(regs)
    return GDataframe(regs, meta)


def check_regs(region_df, chr_name=None, start_name=None, stop_name=None,
               strand_name=None, sample_name=None):
    """ Modifies a region dataframe to be coherent with the GMQL data model

    :param region_df: a pandas Dataframe of regions that is coherent with the GMQL data model
    :param chr_name: (optional) which column of :attr:`~.region_df` is the chromosome
    :param start_name: (optional) which column of :attr:`~.region_df` is the start
    :param stop_name: (optional) which column of :attr:`~.region_df` is the stop
    :param strand_name: (optional) which column of :attr:`~.region_df` is the strand
    :return: a modified pandas Dataframe
    """

    if sample_name is None:
        region_df.index = np.repeat(default_id_sample, len(region_df))
    else:
        region_df = search_column(region_df, id_sample_aliases,
                                  id_sample_types, 'id_sample', sample_name)
        region_df = region_df.set_index("id_sample", drop=True)
        region_df = region_df.sort_index()

    region_df = search_column(region_df,  chr_aliases, chr_types, 'chr', chr_name)
    region_df = search_column(region_df, start_aliases, start_types, 'start', start_name)
    region_df = search_column(region_df, stop_aliases, stop_types, 'stop', stop_name)
    region_df = search_column(region_df, strand_aliases, strand_types, 'strand', strand_name)
    return region_df


def search_column(region_df, names, types, subs, name=None):
    columns = region_df.columns.map(str.lower)
    names = list(map(str.lower, names))

    if name is not None:
        if name.lower() not in columns:
            raise ValueError("{} is not a column of the region dataframe".format(name))
        if check_type(region_df[name], types):
            region_df = region_df.rename(columns={name: subs})
            return region_df
        else:
            raise TypeError("Column {} is not of type {}.".format(name, types))

    isok = False
    for e in columns:
        if e in names and check_type(region_df[e], types):
            region_df = region_df.rename(columns={e: subs})
            isok = True
            break
    if (not isok) and (subs != 'strand'):
        raise ValueError("{} column was not found".format(subs))
    return region_df


def check_type(column, types):
    return column.dtype in types


def check_meta(meta_df, regs_df):
    reg_index = regs_df.index
    reg_samples = set(reg_index.unique())

    meta_index = meta_df.index
    meta_samples = set(meta_index.unique())

    if reg_samples != meta_samples:
        return False
    else:
        return True


def empty_meta(regs):
    index = regs.index.unique()
    df = pd.DataFrame(index=index)
    df['file'] = pd.Series(index=index, data=list(map(lambda x: [str(x)], index)))
    return df


def to_gmql_regions(regs):
    cols = ['chr', 'start', 'stop']
    if 'strand' in regs.columns:
        cols.append('strand')
    cols.extend([c for c in regs.columns if c not in cols])
    return regs[cols]



from .RegionParser import RegionParser
from . import COORDS_DEFAULT
from . import TAB


class BedParser(RegionParser):
    """ Standard Full BED Parser of 10 Columns
    """
    def __init__(self):
        super().__init__(chrPos=0,
                         startPos=1,
                         stopPos=2,
                         strandPos=5,
                         otherPos=[(3, "3", "string"),
                                   (4, "4", 'double'),
                                   (6, "6", 'double'),
                                   (7, "7", 'double'),
                                   (8, "8", 'double'),
                                   (9, "9", "double")],
                         delimiter="\t",
                         coordinate_system=COORDS_DEFAULT,
                         schema_format=TAB)


class ANNParser(RegionParser):
    """ Annotation Parser, 6 columns
    """
    def __init__(self):
        super().__init__(chrPos=0,
                         startPos=1,
                         stopPos=2,
                         strandPos=5,
                         otherPos=[(3, "name", "string"),
                                   (4, "score", 'double')],
                         delimiter="\t",
                         coordinate_system=COORDS_DEFAULT,
                         schema_format=TAB)


class BroadProjParser(RegionParser):
    def __init__(self):
        super().__init__(chrPos=0,
                         startPos=1,
                         stopPos=2,
                         strandPos=None,
                         otherPos=[(3, "name", "string")],
                         delimiter="\t",
                         coordinate_system=COORDS_DEFAULT,
                         schema_format=TAB)


class BasicParser(RegionParser):
    """ Parser for Chr, Start, Stop only (no Strand)
    """
    def __init__(self):
        super().__init__(chrPos=0,
                         startPos=1,
                         stopPos=2,
                         strandPos=None,
                         otherPos=None,
                         delimiter="\t",
                         coordinate_system=COORDS_DEFAULT,
                         schema_format=TAB)


class NarrowPeakParser(RegionParser):
    """ Narrow Peaks Parser. 10 columns
    """
    def __init__(self):
        super().__init__(chrPos=0,
                         startPos=1,
                         stopPos=2,
                         strandPos=5,
                         otherPos=[(3, "name", "string"),
                                   (4, "score", 'double'),
                                   (6, "signalValue", 'double'),
                                   (7, "pValue", 'double'),
                                   (8, "qValue", 'double'),
                                   (9, "peak", "double")],
                         delimiter="\t",
                         coordinate_system=COORDS_DEFAULT,
                         schema_format=TAB)


class RnaSeqParser(RegionParser):
    """ Standard Full BED Parser of 10 Columns
    """
    def __init__(self):
        super().__init__(chrPos=0,
                         startPos=1,
                         stopPos=2,
                         strandPos=3,
                         otherPos=[(4, "name", "string"),
                                   (5, "score", 'double')],
                         delimiter="\t",
                         coordinate_system=COORDS_DEFAULT,
                         schema_format=TAB)


class BedScoreParser(RegionParser):
    """ Standard Full BED Parser of 10 Columns
    """
    def __init__(self):
        super().__init__(chrPos=0,
                         startPos=1,
                         stopPos=2,
                         strandPos=None,
                         otherPos=[(3, "score", 'double')],
                         delimiter="\t",
                         coordinate_system=COORDS_DEFAULT,
                         schema_format=TAB)




from ... import get_python_manager


class GenometricCondition(object):
    def __init__(self, condition_name, argument=None):
        self.condition_name = condition_name
        self.argument = argument
        self.gen_condition = get_python_manager()   \
                            .getOperatorManager() \
                            .getGenometricCondition(self.condition_name, str(self.argument))

    def get_condition_name(self):
        return self.condition_name

    def get_argument(self):
        return self.argument

    def get_gen_condition(self):
        return self.gen_condition


class DLE(GenometricCondition):
    """ Denotes the less-equal distance clause, which selects all the regions of
     the experiment such that their distance from the anchor region is less than, 
     or equal to, N bases. There are two special less-equal distances clauses: 
     DLE(-1) searches for regions of the experiment which overlap with the anchor 
     region (regardless the extent of the overlap), while DLE(0) searched for experiment 
     regions adjacent to, or overlapping, the anchor region
     """
    def __init__(self, limit):
        super(DLE, self).__init__(condition_name="DLE", argument=limit)


class DGE(GenometricCondition):
    """ Greater-equal distance clause, which selects all the regions of the experiment 
    such that their distance from the anchor region is greater than, or equal to, N bases"""
    def __init__(self, limit):
        super(DGE, self).__init__(condition_name="DGE", argument=limit)


class DL(GenometricCondition):
    """ Less than distance clause, which selects all the regions of the experiment
    such that their distance from the anchor region is less than N bases"""
    def __init__(self, limit):
        super(DL, self).__init__(condition_name="DL", argument=limit)


class DG(GenometricCondition):
    """ Greater than distance clause, which selects all the regions of the experiment
    such that their distance from the anchor region is greater than N bases"""
    def __init__(self, limit):
        super(DG, self).__init__(condition_name="DG", argument=limit)


class MD(GenometricCondition):
    """ Denotes the minimum distance clause, which selects the first K regions of an 
    experiment sample at minimal distance from an anchor region of an anchor dataset 
    sample. In case of ties (i.e., regions at the same distance from the anchor region), 
    all tied experiment regions are kept in the result, even if they would exceed 
    the limit of K
    """
    def __init__(self, number):
        super(MD, self).__init__(condition_name="MD", argument=number)


class UP(GenometricCondition):
    """ Upstream."""
    def __init__(self):
        super(UP, self).__init__(condition_name="UP")


class DOWN(GenometricCondition):
    """ Downstream."""
    def __init__(self):
        super(DOWN, self).__init__(condition_name="DOWN")



import os
import xml.etree.ElementTree as ET
from ..FileManagment.SessionFileManager import get_user_dir
from xml.dom import minidom

SESSIONS_TAG = "sessions"
SESSION_TAG = "session"
ADDRESS_TAG = "address"
LAST_USAGE_TAG = "last_usage"
TOKEN_TAG = "token"
TYPE_TAG = "type"

session_file_path = os.path.join(get_user_dir(), "sessions.xml")


class SessionManager:
    def __init__(self, sessions):
        self.sessions = sessions

    def add_session(self, address, auth_token, time, type):
        self.sessions[address] = (time, auth_token, type)

    def remove_session(self, address):
        self.sessions.pop(address)

    def get_session(self, address=None):
        # if there is no session
        if len(self.sessions) == 0:
            return None
        else:
            if address is None:
                # get last session
                ordered_sessions = sorted(self.sessions.items(), key=lambda x: x[1][0])
                last_session = ordered_sessions[0]
                address = last_session[0]
                auth_token = last_session[1][1]
                s_type = last_session[1][2]
                return address, auth_token, s_type
            else:
                res = self.sessions.get(address)
                if res is None:
                    return None
                else:
                    return address, res[1], res[2]


def load_sessions():
    if not os.path.isfile(session_file_path):
        return SessionManager(dict())
    tree = ET.ElementTree(file=session_file_path)
    sessions = tree.getroot()
    res = {}
    for session_node in sessions.findall(SESSION_TAG):
        sn_address = session_node.find(ADDRESS_TAG).text
        sn_auth_token = session_node.find(TOKEN_TAG).text
        sn_last_usage = int(session_node.find(LAST_USAGE_TAG).text)
        sn_type = session_node.find(TYPE_TAG).text
        res[sn_address] = (sn_last_usage, sn_auth_token, sn_type)

    return SessionManager(res)


def store_sessions(sessions):
    sessions_node = ET.Element(SESSIONS_TAG)
    for address in sessions.keys():
        time, auth_token, s_type = sessions[address]
        session_node = ET.SubElement(sessions_node, SESSION_TAG)

        address_node = ET.SubElement(session_node, ADDRESS_TAG)
        address_node.text = address

        at_node = ET.SubElement(session_node, TOKEN_TAG)
        at_node.text = auth_token

        time_node = ET.SubElement(session_node, LAST_USAGE_TAG)
        time_node.text = str(time)

        type_node = ET.SubElement(session_node, TYPE_TAG)
        type_node.text = s_type

    # tree = ET.ElementTree(sessions_node)
    rough_string = ET.tostring(sessions_node, encoding="utf-8")
    reparsed = minidom.parseString(rough_string)
    pretty_string = reparsed.toprettyxml()
    with open(session_file_path, "w") as f:
        f.write(pretty_string)



"""
Setting up the environment of the library
"""
import logging
import time
import os
import sys
import atexit
from sys import stdout
from py4j.java_gateway import JavaGateway, launch_gateway, GatewayParameters
from py4j.java_collections import ListConverter
from .FileManagment import TempFileManager
from pkg_resources import resource_filename
from .RemoteConnection.SessionManager import load_sessions, store_sessions
from .FileManagment.SessionFileManager import initialize_user_folder
from .FileManagment.DependencyManager import DependencyManager

"""
    Version management
"""


def get_version():
    version_file_name = resource_filename("gmql", os.path.join("resources", "version"))
    with open(version_file_name, "r") as f_ver:
        version = f_ver.read().strip()
    return version


__version__ = get_version()


def __get_github_url():
    github_fn = resource_filename("gmql", os.path.join("resources", "github_url"))
    with open(github_fn, "r") as f_ver:
        url = f_ver.read().strip()
    return url


__gmql_jar = "pythonAPI-{}.jar".format(__version__)
__backend_download_url = __get_github_url()

"""
    Logging and progress bars
"""


def set_logger(logger_name):
    """
    The library has a personal logger that tells the user the intermediate steps of the
    commands that he/she calls
    :param: name of the logger
    :return: the setup logger
    """
    logger = logging.getLogger(logger_name)
    logger.setLevel(logging.INFO)

    # create a stream handler
    handler = logging.StreamHandler(stdout)
    handler.setLevel(logging.INFO)

    # create a logging format
    formatter = logging.Formatter(
        '%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    handler.setFormatter(formatter)

    # add the handlers to the logger
    logger.addHandler(handler)
    return logger


__disable_progress = False


def set_progress(how):
    """ Enables or disables the progress bars for the loading, writing and downloading
    of datasets

    :param how: True if you want the progress bar, False otherwise
    :return: None

    Example::

        import gmql as gl

        gl.set_progress(True)   # abilitates progress bars
        # ....do something...
        gl.set_progress(False)  # removes progress bars
        # ....do something...
    """
    global __disable_progress
    if isinstance(how, bool):
        __disable_progress = ~how
    else:
        raise ValueError(
            "how must be a boolean. {} was found".format(type(how)))


def is_progress_enabled():
    return not __disable_progress


def set_logging(how):
    """ Enables or disables the logging mechanism of PyGMQL.

    :param how: True if you want the logging, False otherwise
    :return: None

    Example::

        import gmql as gl

        gl.set_logging(True)   # abilitates the logging
        # ....do something...
        gl.set_logging(False)  # removes the logging
        # ....do something...
    """
    if isinstance(how, bool):
        if how:
            logging.disable(logging.NOTSET)
        else:
            logging.disable(logging.CRITICAL)
    else:
        raise ValueError(
            "how must be a boolean. {} was found".format(type(how)))


_metadata_profiling = True


def set_meta_profiling(how):
    """ Enables or disables the profiling of metadata at the loading of a GMQLDataset

    :param how: True if you want to analyze the metadata when a GMQLDataset is created
                by a load_from_*. False otherwise. (Default=True)
    :return: None
    """
    global _metadata_profiling
    if isinstance(how, bool):
        _metadata_profiling = how
    else:
        raise TypeError("how must be boolean. {} was provided".format(type(how)))


"""
    Backend management
"""


def __get_python_api_package(gateway):
    return gateway.jvm.it.polimi.genomics.pythonapi


def __start_gmql_manager(python_api_package):
    pythonManager = python_api_package.PythonManager
    pythonManager.startEngine()
    return pythonManager


def _get_gateway():
    global gateway

    if gateway is None:
        # Starting the GMQL manager
        start()
        return gateway
    else:
        return gateway


def get_python_manager():
    global pythonManager

    if pythonManager is None:
        # Starting the GMQL manager
        start()
        return pythonManager
    else:
        return pythonManager


"""
    GMQL Logger configuration
"""
logger_name = "PyGMQL logger"
logger = set_logger(logger_name)

"""
    Initializing the JVM with the 
"""
gateway, pythonManager = None, None


def start():
    global pythonManager, gateway, __dependency_manager

    java_home = os.environ.get("JAVA_HOME")
    if java_home is None:
        raise SystemError("The environment variable JAVA_HOME is not set")
    java_path = os.path.join(java_home, "bin", "java")
    gmql_jar_fn = __dependency_manager.resolve_dependencies()
    _port = launch_gateway(classpath=gmql_jar_fn, die_on_exit=True,
                           java_path=java_path, javaopts=['-Xmx4096m'])
    gateway = JavaGateway(gateway_parameters=GatewayParameters(port=_port,
                                                               auto_convert=True))
    python_api_package = __get_python_api_package(gateway)
    pythonManager = __start_gmql_manager(python_api_package)

    if sys.platform.startswith("win32"):
        # if we are on windows set the hadoop home to winutils.exe
        hadoopFolder_fn = resource_filename(
            "gmql", os.path.join("resources", "hadoop")
        )
        pythonManager.setHadoopHomeDir(hadoopFolder_fn)

    # setting the spark tmp folder
    pythonManager.setSparkLocalDir(folders['spark'])


def stop():
    global gateway, __session_manager
    # flushing the tmp files
    TempFileManager.flush_everything()
    store_sessions(__session_manager.sessions)
    if gateway is not None:
        gateway.shutdown()


atexit.register(stop)


class GMQLManagerNotInitializedError(Exception):
    pass


# Setting up the temporary files folder
folders = TempFileManager.initialize_tmp_folders()

"""
    LANGUAGE CONVERSIONS
"""


def to_java_list(l):
    ListConverter().convert(l, gateway)


def none():
    return get_python_manager().getNone()


def Some(thing):
    return get_python_manager().getSome(thing)


"""
    Remote manager management
"""

remote_manager = None
remote_address = None

__session_manager = None
__session_type = None
__access_time = None
__dependency_manager = None


def __initialize_dependency_manager():
    global __dependency_manager
    __dependency_manager = DependencyManager()


def __initialize_session_manager():
    global __session_manager
    __session_manager = load_sessions()


def get_remote_manager():
    """ Returns the current remote manager

    :return: a RemoteManager
    """
    global remote_manager
    return remote_manager


def get_session_manager():
    global __session_manager
    return __session_manager


def login():
    """ Enables the user to login to the remote GMQL service.
    If both username and password are None, the user will be connected as guest.
    """
    global remote_manager, remote_address, __session_manager
    res = __session_manager.get_session(remote_address)
    if res is None:
        # there is no session for this address, let's login as guest
        rm = RemoteManager(address=remote_address)
        rm.login()
        session_type = "guest"
    else:
        # there is a previous session for this address, let's do an auto login
        # using that access token
        rm = RemoteManager(address=remote_address, auth_token=res[1])
        # if the access token is not valid anymore (therefore we are in guest mode)
        # the auto_login function will perform a guest login from scratch
        session_type = rm.auto_login(how=res[2])
    # store the new session
    remote_manager = rm
    access_time = int(time.time())
    auth_token = rm.auth_token
    __session_manager.add_session(remote_address, auth_token, access_time, session_type)


def set_remote_address(address):
    """ Enables the user to set the address of the GMQL remote service

    :param address: a string representing the URL of GMQL remote service
    :return: None
    """
    global remote_address
    remote_address = address


def logout():
    """ The user can use this command to logout from the remote service

    :return: None
    """
    global remote_manager
    remote_manager.logout()


def execute_remote():
    global remote_manager
    remote_manager.execute_remote_all()


"""
    Execution mode management
"""

mode = "local"


def set_mode(how):
    """ Sets the behavior of the API

    :param how: if 'remote' all the execution is performed on the remote server; if 'local' all
           it is executed locally. Default = 'local'
    :return: None
    """
    global mode
    if how == "local":
        mode = how
    elif how == "remote":
        mode = how
    else:
        raise ValueError("how must be 'local' or 'remote'")


def get_mode():
    global mode
    return mode


"""
    Datasets management
"""

__source_table = None


def __initialize_source_table():
    global __source_table
    from .dataset.loaders.Sources import SourcesTable
    __source_table = SourcesTable()


def _get_source_table():
    global __source_table
    return __source_table


"""
    EXPOSING INTERNAL FEATURES
"""
from .dataset.GDataframe import from_pandas
from .dataset.loaders.Loader import load_from_path, load_from_remote, load
from .dataset import parsers
# the possible aggregations
from .dataset.DataStructures.Aggregates import *
from .dataset.DataStructures.ExpressionNodes import *
# the possible join conditions
from .dataset.DataStructures.GenometricPredicates import *
# for interacting with the remote cluster
from .RemoteConnection.RemoteManager import RemoteManager

# modules for the machine learning library
# from . import ml


# __initialize_remote_manager()
__initialize_source_table()
initialize_user_folder()
__initialize_session_manager()
__initialize_dependency_manager()